@@ -4,7 +4,7 @@ from nptyping import Float, NDArray
 from dataclasses import dataclass
 from typing import Any, Dict, List, Optional, Union
 from pyspark.sql import Column, DataFrame
-from pyspark.sql.types import StringType, StructField
+from pyspark.sql.types import StringType, StructField, IntegerType
 from scipy import stats
 from typeguard import typechecked
 from . import functions as gwas_fx
@@ -22,11 +22,12 @@ def linear_regression(genotype_df: DataFrame,
                       contigs: Optional[List[str]] = None,
                       add_intercept: bool = True,
                       values_column: Union[str, Column] = 'values',
-                      dt: type = np.float64) -> DataFrame:
+                      dt: type = np.float64,
+                      verbose_output: bool = False) -> DataFrame:
     '''
     Uses linear regression to test for association between genotypes and one or more phenotypes.
     The implementation is a distributed version of the method used in regenie: 
-    https://www.biorxiv.org/content/10.1101/2020.06.19.162354v2
+    https://www.nature.com/articles/s41588-021-00870-7
 
     Implementation details:
 
@@ -78,6 +79,10 @@ def linear_regression(genotype_df: DataFrame,
                         ``genotype_df`` should have a column with this name and a numeric array type. If a column
                         expression is provided, the expression should return a numeric array type.
         dt : The numpy datatype to use in the linear regression test. Must be ``np.float32`` or ``np.float64``.
+        verbose_output: Whether or not to generate additional test statistics (n, sum_x, y_transpose_x)
+                        to the output DataFrame.  These values are derived directly from phenotype_df and genotype_df,
+                        and does not reflect any standardization performed as part of the implementation of
+                        linear_regression.
 
     Returns:
         A Spark DataFrame that contains
@@ -88,6 +93,10 @@ def linear_regression(genotype_df: DataFrame,
         - ``tvalue``: The T statistic
         - ``pvalue``: P value estimated from a two sided T-test
         - ``phenotype``: The phenotype name as determined by the column names of ``phenotype_df``
+        - ``n``(int): (verbose_output only) number of samples with non-null phenotype
+        - ``sum_x``(float): (verbose_output only) sum of genotype inputs
+        - ``y_transpose_x``(float): (verbose_output only) dot product of phenotype response (missing values encoded as zeros)
+                             and genotype input, i.e. phenotype value * number of alternate alleles
     '''
 
     gwas_fx._check_spark_version(genotype_df.sql_ctx.sparkSession)
@@ -98,16 +107,6 @@ def linear_regression(genotype_df: DataFrame,
 
     genotype_df = gwas_fx._prepare_genotype_df(genotype_df, values_column, sql_type)
 
-    # Construct output schema
-    result_fields = [
-        StructField('effect', sql_type),
-        StructField('stderror', sql_type),
-        StructField('tvalue', sql_type),
-        StructField('pvalue', sql_type),
-        StructField('phenotype', StringType())
-    ]
-    result_struct = gwas_fx._output_schema(genotype_df.schema.fields, result_fields)
-
     C = covariate_df.to_numpy(dt, copy=True)
     if add_intercept:
         C = gwas_fx._add_intercept(C, phenotype_df.shape[0])
@@ -117,6 +116,7 @@ def linear_regression(genotype_df: DataFrame,
     Y = phenotype_df.to_numpy(dt, copy=True)
     Y_mask = (~np.isnan(Y)).astype(dt)
     Y = np.nan_to_num(Y, copy=False)
+    Y_for_verbose_output = np.copy(Y) if verbose_output else None
     Y -= Y.mean(axis=0)  # Mean-center
     Y = gwas_fx._residualize_in_place(Y, Q) * Y_mask  # Residualize
     Y_scale = np.sqrt(np.sum(Y**2, axis=0) / (Y_mask.sum(axis=0) - Q.shape[1]))
@@ -126,13 +126,8 @@ def linear_regression(genotype_df: DataFrame,
 
     dof = C.shape[0] - C.shape[1] - 1
 
-    def map_func(pdf_iterator):
-        for pdf in pdf_iterator:
-            yield gwas_fx._loco_dispatch(pdf, Y_state, _linear_regression_inner, Y_mask, Y_scale, Q,
-                                         dof,
-                                         phenotype_df.columns.to_series().astype('str'))
-
-    return genotype_df.mapInPandas(map_func, result_struct)
+    return _generate_linreg_output(genotype_df, sql_type, Y_state, Y_mask, Y_scale, Q, dof,
+                                   phenotype_df, Y_for_verbose_output, verbose_output)
 
 
 @dataclass
@@ -173,8 +168,9 @@ def _create_one_YState(Y: NDArray[(Any, Any), Float], phenotype_df: pd.DataFrame
 # @typechecked -- typeguard does not support numpy array
 def _linear_regression_inner(genotype_pdf: pd.DataFrame, Y_state: YState,
                              Y_mask: NDArray[(Any, Any), Float], Y_scale: NDArray[(Any, ), Float],
-                             Q: NDArray[(Any, Any), Float], dof: int,
-                             phenotype_names: pd.Series) -> pd.DataFrame:
+                             Q: NDArray[(Any, Any), Float], dof: int, phenotype_names: pd.Series,
+                             Y_raw: Optional[NDArray[(Any, Any), Float]],
+                             verbose_output: Optional[bool]) -> pd.DataFrame:
     '''
     Applies a linear regression model to a block of genotypes. We first project the covariates out of the
     genotype block and then perform single variate linear regression for each site.
@@ -191,6 +187,13 @@ def _linear_regression_inner(genotype_pdf: pd.DataFrame, Y_state: YState,
     '''
 
     X = np.column_stack(genotype_pdf[_VALUES_COLUMN_NAME].array)
+    del genotype_pdf[_VALUES_COLUMN_NAME]
+    num_genotypes = genotype_pdf.shape[0]
+    out_df = pd.concat([genotype_pdf] * Y_state.Y.shape[1])
+    if verbose_output:
+        out_df["n"] = list(Y_mask.sum(axis=0).repeat(num_genotypes))
+        out_df["sum_x"] = list(np.ravel(Y_mask.T @ X))
+        out_df["y_transpose_x"] = list(np.ravel(Y_raw.T @ X))
     X = gwas_fx._residualize_in_place(X, Q)
 
     XdotY = Y_state.Y.T @ X
@@ -200,9 +203,6 @@ def _linear_regression_inner(genotype_pdf: pd.DataFrame, Y_state: YState,
     T = betas / standard_error
     pvalues = 2 * stats.distributions.t.sf(np.abs(T), dof)
 
-    del genotype_pdf[_VALUES_COLUMN_NAME]
-    num_genotypes = genotype_pdf.shape[0]
-    out_df = pd.concat([genotype_pdf] * Y_state.Y.shape[1])
     Y_scale_mat = Y_scale[:, None]
     out_df['effect'] = list(np.ravel(betas * Y_scale_mat))
     out_df['stderror'] = list(np.ravel(standard_error * Y_scale_mat))
@@ -211,3 +211,33 @@ def _linear_regression_inner(genotype_pdf: pd.DataFrame, Y_state: YState,
     out_df['phenotype'] = phenotype_names.repeat(num_genotypes).tolist()
 
     return out_df
+
+
+def _generate_linreg_output(genotype_df, sql_type, Y_state, Y_mask, Y_scale, Q, dof, phenotype_df,
+                            Y_for_verbose_output, verbose_output) -> DataFrame:
+    # Construct output schema
+    result_fields = [
+        StructField('effect', sql_type),
+        StructField('stderror', sql_type),
+        StructField('tvalue', sql_type),
+        StructField('pvalue', sql_type),
+        StructField('phenotype', StringType())
+    ]
+
+    if verbose_output:
+        result_fields += ([
+            StructField('n', IntegerType()),
+            StructField('sum_x', sql_type),
+            StructField('y_transpose_x', sql_type)
+        ])
+
+    result_struct = gwas_fx._output_schema(genotype_df.schema.fields, result_fields)
+
+    def map_func(pdf_iterator):
+        for pdf in pdf_iterator:
+            yield gwas_fx._loco_dispatch(pdf, Y_state, _linear_regression_inner, Y_mask, Y_scale, Q,
+                                         dof,
+                                         phenotype_df.columns.to_series().astype('str'),
+                                         Y_for_verbose_output, verbose_output)
+
+    return genotype_df.mapInPandas(map_func, result_struct)
